package character

import (
	"log"

	functionset "example.com/database/functionset"
	inventoryslot "example.com/database/inventoryslot"
	person "example.com/database/person"
	role "example.com/database/role"

	"gorm.io/gorm"
)

type Character struct {
	gorm.Model
	Name         string                        `gorm:"primaryKey; uniqueIndex" json:"name"`
	Owners       []person.Person               `gorm:"many2many:character_owner_associations" json:"owners"` // Unique
	Roles        []role.Role                   `gorm:"many2many:character_role_associations" json:"roles"`   // Unique
	FunctionSets []functionset.FunctionSet     `gorm:"many2many:character_functionset_associations" json:"function_sets"`
	Inventory    []inventoryslot.InventorySlot `gorm:"many2many:character_inventory_associations" json:"inventory"`
}

func (character Character) Create(db *gorm.DB) error {
	result := db.Create(&character)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func (character Character) getAssociations(db *gorm.DB) {
	db.Model(&character).Association("Owners").Find(&character.Owners)
	db.Model(&character).Association("Roles").Find(&character.Roles)
	db.Model(&character).Association("FunctionSets").Find(&character.FunctionSets)
	db.Model(&character).Association("Inventory").Find(&character.Inventory)
}

func (character *Character) Get(db *gorm.DB, inputCharacter string) {
	db.Where("name = ?", inputCharacter).Take(&character)
	character.getAssociations(db)
}

func (character Character) Update(db *gorm.DB) error {
	var originalCharacter Character
	originalCharacter.Get(db, character.Name)
	ownersError := db.Model(&originalCharacter).Association("Owners").Replace(&character.Owners)
	if ownersError != nil {
		return ownersError
	}
	rolesError := db.Model(&originalCharacter).Association("Roles").Replace(&character.Roles)
	if rolesError != nil {
		return rolesError
	}
	functionSetsError := db.Model(&originalCharacter).Association("FunctionSets").Replace(&character.FunctionSets)
	if functionSetsError != nil {
		return functionSetsError
	}
	inventoryError := db.Model(&originalCharacter).Association("Inventory").Replace(&character.Inventory)
	if inventoryError != nil {
		return inventoryError
	}
	return nil
}

func (character Character) Delete(db *gorm.DB) error {
	result := db.Delete(&character)
	if result.Error != nil {
		return result.Error
	}
	return nil
}

func Create(db *gorm.DB, name string, owners []string, roles []string, functionsets []int, inventory []int) error {
	return Character{
		Name:         name,
		Owners:       *person.Get(db, owners),
		Roles:        *role.Get(db, roles),
		FunctionSets: *functionset.Get(db, functionsets),
		Inventory:    *inventoryslot.Get(db, inventory),
	}.Create(db)
}

func Get(db *gorm.DB, inputCharacters []string) *[]Character {
	var outputCharacters []Character
	for _, inputCharacter := range inputCharacters {
		var outputCharacter Character
		outputCharacter.Get(db, inputCharacter)
		outputCharacters = append(outputCharacters, outputCharacter)
	}
	return &outputCharacters
}

func GetAll(db *gorm.DB) *[]Character {
	var outputCharacterNames []string
	result := db.Model(&Character{}).Select("name").Find(&outputCharacterNames)
	if result.Error != nil {
		log.Println(result.Error)
	}
	return Get(db, outputCharacterNames)
}

func Update(db *gorm.DB, name string, owners []string, roles []string, functionsets []int, inventory []int) error {
	return Character{
		Name:         name,
		Owners:       *person.Get(db, owners),
		Roles:        *role.Get(db, roles),
		FunctionSets: *functionset.Get(db, functionsets),
		Inventory:    *inventoryslot.Get(db, inventory),
	}.Update(db)
}

func Delete(db *gorm.DB, inputCharacters []string) {
	characters := Get(db, inputCharacters)
	for _, character := range *characters {
		character.Delete(db)
	}
}
